# Shared functionality to operate with Wannier programs (Wannier.jl and Wannier90).
# The communication happens with the Wannier90 file formats which Wannier.jl also understands.

using Dates

"""
A Gaussian-shaped initial guess. Can be used as an approximation of an s- or σ-like orbital.
"""
struct GaussianWannierProjection
    center::AbstractVector
end

function (proj::GaussianWannierProjection)(basis::PlaneWaveBasis, ps)
    # associate a center with the fourier transform of the corresponding gaussian
    # TODO: what is the normalization here?

    model = basis.model
    map(ps) do p
        p_cart = recip_vector_red_to_cart(model, p)
        exp( 2π*(-im*dot(p, proj.center) - dot(p_cart, p_cart) / 4) )
    end
end

@doc raw"""
A hydrogenic initial guess.

`α` is the diffusivity, ``\frac{Z}{a}`` where ``Z`` is the atomic number and
    ``a`` is the Bohr radius.
"""
struct HydrogenicWannierProjection
    center::AbstractVector
    n::Integer
    l::Integer
    m::Integer
    α::Real
end

function (proj::HydrogenicWannierProjection)(basis::PlaneWaveBasis, ps)
    # TODO: Performance can probably be improved a lot here.
    # TODO: Some of this logic could be used for the pswfc from the UPF PSPs.

    # this is same as QE pw2wannier90, r = exp(x) / α
    xmin = -6.0
    dx = 0.025
    rmax = 10.0
    n_r = round(Int, (log(rmax) - xmin) / dx) + 1
    x = range(; start=xmin, length=n_r, step=dx)
    # rgrid depends on α
    r = exp.(x) ./ proj.α
    dr = r .* dx
    R = radial_hydrogenic(r, proj.n, proj.α)
    r2_R_dr = r.^2 .* R .* dr

    model = basis.model
    map(ps) do p
        p_cart = recip_vector_red_to_cart(model, p)
        pnorm = norm(p_cart)

        radial_part = 0.0
        for (ir, rval) in enumerate(r)
            @inbounds radial_part += r2_R_dr[ir] * sphericalbesselj_fast(proj.l, pnorm * rval)
        end

        # both p and proj.center in reduced coordinates
        center_offset = exp(-im*2π*dot(p, proj.center))
        # without the 4π normalization because the matrix will be orthonormalized anyway
        angular_part = ylm_real(proj.l, proj.m, p_cart) * (-im)^proj.l

        center_offset * angular_part * radial_part
    end
end

"""
Write a win file at the indicated prefix.
Parameters to Wannier90 can be added as kwargs: e.g. `num_iter=500`.
"""
function write_w90_win(fileprefix::String, basis::PlaneWaveBasis;
                       bands_plot=false, wannier_plot=false, kwargs...)
    @assert :num_bands in keys(kwargs)  # Required parameter
    @assert :num_wann  in keys(kwargs)  # Required parameter
    basis.kgrid isa MonkhorstPack || error("The basis must be constructed from a MP grid.")

    open(fileprefix *".win", "w") do fp
        println(fp, "! Generated by DFTK.jl at $(now())\n")

        # Drop parameters
        for (key, value) in pairs(kwargs)
            @printf fp "%-20s =   %-30s\n" key value
        end
        if wannier_plot
            println(fp, "wvfn_formatted = True")
            println(fp, "wannier_plot   = True")
        end

        # Delimiter for system
        println(fp, "\n", "!"^20 * " System \n\n")

        # Lattice vectors are in rows in Wannier90
        println(fp, "begin unit_cell_cart\nbohr")
        for vec in eachcol(basis.model.lattice)
            @printf fp "%10.6f %10.6f %10.6f \n" vec...
        end
        println(fp, "end unit_cell_cart \n")

        println(fp, "begin atoms_frac")
        for (element, position) in zip(basis.model.atoms, basis.model.positions)
            @printf fp "%-2s %10.6f %10.6f %10.6f \n" element.symbol position...
        end
        println(fp, "end atoms_frac\n")

        # Delimiter for k-points block
        println(fp, "!"^20 * " k_points\n")

        if bands_plot
            kpath  = irrfbz_path(basis.model)
            length(kpath.paths) > 1 || @warn(
                "Only first kpath branch considered in write_w90_win")
            path = kpath.paths[1]

            println(fp, "begin kpoint_path")
            for i = 1:length(path)-1
                A, B = path[i:i+1]  # write segment A -> B
                @printf(fp, "%s %10.6f %10.6f %10.6f  ", A, round.(kpath.points[A], digits=5)...)
                @printf(fp, "%s %10.6f %10.6f %10.6f\n", B, round.(kpath.points[B], digits=5)...)
            end
            println(fp, "end kpoint_path")
            println(fp, "bands_plot = true\n")
        end

        println(fp, "mp_grid : $(basis.kgrid.kgrid_size[1]) $(basis.kgrid.kgrid_size[2]) ",
                "$(basis.kgrid.kgrid_size[3])\n")
        println(fp, "begin kpoints")
        for kpt in basis.kpoints
            @printf fp  "%10.10f %10.10f %10.10f\n" kpt.coordinate...
        end
        println(fp, "end kpoints")
    end
    nothing
end


"""
Read the .nnkp file provided by the preprocessing routine of Wannier90
(i.e. "wannier90.x -pp prefix")
Returns:
1) the array 'nnkpts' of k points, their respective nearest neighbors and associated
   shifing vectors (non zero if the neighbor is located in another cell).
2) the number 'nntot' of neighbors per k point.

TODO: add the possibility to exclude bands
"""
function read_w90_nnkp(fileprefix::String)
    fn = fileprefix * ".nnkp"
    !isfile(fn) && error("Expected file $fileprefix.nnkp not found.")

    # Extract nnkp block
    lines = readlines(fn)
    ibegin_nnkp = findfirst(lines .== "begin nnkpts")
    iend_nnkp   = findfirst(lines .== "end nnkpts")
    @assert !isnothing(ibegin_nnkp) && !isnothing(iend_nnkp)

    nntot  = parse(Int, lines[ibegin_nnkp + 1])
    nnkpts = map(ibegin_nnkp+2:iend_nnkp-1) do iline
        splitted = parse.(Int, split(lines[iline], ' ', keepempty=false))
        @assert length(splitted) == 5

        # Meaning of the entries:
        # 1st: Index of k-point
        # 2nd: Index of periodic image of k+b k-point
        # 3rd: Shift vector to get k-point of ik_plus_b to the actual k+b point required
        (; ik=splitted[1], ik_plus_b=splitted[2], G_shift=splitted[3:5])
    end
    (; nntot, nnkpts)
end


"""
Write the eigenvalues in a format readable by Wannier90.
"""
function write_w90_eig(fileprefix::String, eigenvalues; n_bands)
    open("$fileprefix.eig", "w") do fp
        for (k, εk) in enumerate(eigenvalues)
            for (n, εnk) in enumerate(εk[1:n_bands])
                @printf fp "%3i  %3i   %25.18f \n" n k auconvert(Unitful.eV, εnk).val
            end
        end
    end
    nothing
end


@timing function write_w90_unk(fileprefix::String, basis, ψ; n_bands, spin=1)
    fft_size = basis.fft_size

    for ik in krange_spin(basis, spin)
        open(dirname(fileprefix) * (@sprintf "/UNK%05i.%i" ik spin), "w") do fp
            println(fp, "$(fft_size[1]) $(fft_size[2]) $(fft_size[3]) $ik $n_bands")
            for iband = 1:n_bands
                ψnk_real = ifft(basis, basis.kpoints[ik], @view ψ[ik][:, iband])
                for iz = 1:fft_size[3], iy = 1:fft_size[2], ix = 1:fft_size[1]
                    @printf fp "%25.18f %25.18f\n" real(ψnk_real[ix, iy, iz]) imag(ψnk_real[ix, iy, iz])
                end
            end
        end
    end
    nothing
end


@doc raw"""
Computes the matrix ``[M^{k,b}]_{m,n} = \langle u_{m,k} | u_{n,k+b} \rangle``
for given `k`.

`G_shift` is the "shifting" vector, correction due to the periodicity conditions
imposed on ``k \to  ψ_k``.
It is non zero if `k_plus_b` is taken in another unit cell of the reciprocal lattice.
We use here that:
``u_{n(k + G_{\rm shift})}(r) = e^{-i*\langle G_{\rm shift},r \rangle} u_{nk}``.
"""
@views function overlap_Mmn_k_kpb(basis::PlaneWaveBasis, ψ, ik, ik_plus_b, G_shift, n_bands)
    # Search for common Fourier modes and their resp. indices in Bloch states k and k+b
    # TODO Check if this can be improved using the G vector mapping in the kpoints
    k        = basis.kpoints[ik]
    k_plus_b = basis.kpoints[ik_plus_b]
    equivalent_G_vectors = [(ip, index_G_vectors(basis, k_plus_b, p + G_shift))
                            for (ip, p) in enumerate(G_vectors(basis, k))]
    ip        = [eqG[1] for eqG in equivalent_G_vectors if !isnothing(eqG[2])]
    ip_plus_b = [eqG[2] for eqG in equivalent_G_vectors if !isnothing(eqG[2])]

    # Compute overlaps
    # TODO This should be improved ...
    Mkb = zeros(ComplexF64, (n_bands, n_bands))
    for n = 1:n_bands
        for m = 1:n_bands
            # Select the coefficient in right order
            Mkb[m, n] = dot(ψ[ik][ip, m], ψ[ik_plus_b][ip_plus_b, n])
        end
    end
    iszero(Mkb) && return Matrix(I, n_bands, n_bands)
    Mkb
end

@timing function write_w90_mmn(fileprefix::String, basis::PlaneWaveBasis, ψ, nnkp; n_bands)
    open(fileprefix * ".mmn", "w") do fp
        println(fp, "Generated by DFTK at $(now())")
        println(fp, "$n_bands  $(length(ψ))  $(nnkp.nntot)")
        for (ik, ik_plus_b, G_shift) in nnkp.nnkpts
            @printf fp "%i  %i  %i  %i  %i \n" ik ik_plus_b G_shift...
            for ovlp in overlap_Mmn_k_kpb(basis, ψ, ik, ik_plus_b, G_shift, n_bands)
                @printf fp "%22.18f %22.18f \n" real(ovlp) imag(ovlp)
            end
        end
    end
    nothing
end

@doc raw"""
Compute the starting matrix for Wannierization.

Wannierization searches for a unitary matrix ``U_{m_n}``.
As a starting point for the search, we can provide an initial guess function ``g``
for the shape of the Wannier functions, based on what we expect from knowledge of the problem or physical intuition.
This starting matrix is called ``[A_k]_{m,n}``, and is computed as follows:
``[A_k]_{m,n} = \langle ψ_m^k | g^{\text{per}}_n \rangle``.
The matrix will be orthonormalized by the chosen Wannier program, we don't need to do so ourselves.

Centers are to be given in lattice coordinates and G_vectors in reduced coordinates.
The dot product is computed in the Fourier space.

Given an orbital ``g_n``, the periodized orbital is defined by :
 ``g^{per}_n =  \sum\limits_{R \in {\rm lattice}} g_n( \cdot - R)``.
The Fourier coefficient of ``g^{per}_n`` at any G
is given by the value of the Fourier transform of ``g_n`` in G.

Each projection is a callable object that accepts the basis and some p-points as an argument,
and returns the Fourier transform of ``g_n`` at the p-points.
"""
function compute_amn_kpoint(basis::PlaneWaveBasis, kpt, ψk, projections, n_bands)
    n_wannier = length(projections)
    # TODO This function should be improved in performance

    ps = vec(map(G -> G .+ kpt.coordinate, G_vectors(basis)))  # all p = k+G in reduced coordinates
    Ak = zeros(eltype(ψk), (n_bands, n_wannier))

    # Compute Ak
    for n = 1:n_wannier
        proj = projections[n]
        gn_per = proj(basis, ps[kpt.mapping])
        # Fourier coeffs of gn_per for k+G in common with ψk
        # Functions are l^2 normalized in Fourier, in DFTK conventions.
        coeffs_gn_per = gn_per ./ norm(gn_per)
        # Compute overlap
        for m = 1:n_bands
            Ak[m, n] = dot(ψk[:, m], coeffs_gn_per)
        end
    end
    Ak
end


@timing function write_w90_amn(
        fileprefix::String,
        basis::PlaneWaveBasis,
        projections,
        ψ;
        n_bands)
    open(fileprefix * ".amn", "w") do fp
        println(fp, "Generated by DFTK at $(now())")
        println(fp, "$n_bands   $(length(basis.kpoints))  $(length(projections))")

        for (ik, (ψk, kpt)) in enumerate(zip(ψ, basis.kpoints))
            Ak = compute_amn_kpoint(basis, kpt, ψk, projections, n_bands)
            for n = 1:size(Ak, 2)
                for m = 1:size(Ak, 1)
                    @printf(fp, "%3i %3i %3i  %22.18f %22.18f \n",
                            m, n, ik, real(Ak[m, n]), imag(Ak[m, n]))
                end
            end
        end
    end
    nothing
end

"""
Default random Gaussian guess for maximally-localised wannier functions
generated in reduced coordinates.
"""
default_wannier_centers(n_wannier) = [GaussianWannierProjection(rand(3)) for _ = 1:n_wannier]

"""
Shared file writing code for Wannier.jl and Wannier90.
"""
@timing function write_wannier90_files(preprocess_call, scfres;
        n_bands,
        n_wannier,
        projections,
        fileprefix,
        wannier_plot,
        kwargs...)
    # TODO None of the routines consider spin at the moment
    @assert scfres.basis.model.spin_polarization in (:none, :spinless)
    @assert length(projections) == n_wannier

    # TODO Use band_data_to_dict to get this easily MPI compatible.

    # Undo symmetry operations to get full k-point list
    scfres_unfold = unfold_bz(scfres)
    basis = scfres_unfold.basis
    ψ = scfres_unfold.ψ

    # Make wannier directory ...
    mkpath(dirname(fileprefix))

    # Write input file and launch preprocessing
    write_w90_win(fileprefix, basis;
                  num_wann=n_wannier, num_bands=n_bands, wannier_plot, kwargs...)
    nnkp = preprocess_call()

    # Files for main wannierization run
    write_w90_eig(fileprefix, scfres_unfold.eigenvalues; n_bands)
    write_w90_amn(fileprefix, basis, projections, ψ; n_bands)
    write_w90_mmn(fileprefix, basis, ψ, nnkp; n_bands)

    # Writing the unk files is expensive (requires FFTs), so only do if needed.
    if wannier_plot
        write_w90_unk(fileprefix, basis, ψ; n_bands, spin=1)
    end
end
